/*
  Test.h - Test library for Wiring - implementation
  Copyright (c) 2006 John Doe.  All right reserved.
*/


// include this library's description file
#include "PropWiFiESP32.h"


// Constructor /////////////////////////////////////////////////////////////////
// Function that handles the creation and setup of instances




void PropWiFiESP32::enableHeartbeating(bool state)
{
	mEnableHeartbeating = state;
}
PropWiFiESP32 *reference = NULL;
void WiFiEvent(WiFiEvent_t event){
	if (reference != NULL)
	{
		reference->ClassWiFiEvent(event);
	}
}

void PropWiFiESP32::ClassWiFiEvent(WiFiEvent_t event){
	IPAddress getMyIP;
	IPAddress subnetMask;
    switch(event) {
      case SYSTEM_EVENT_STA_GOT_IP:
          //When connected set 
          //initializes the UDP state
          //This initializes the transfer buffer
		  getMyIP = WiFi.localIP();
		  subnetMask = WiFi.subnetMask();
		  mBCast[0] = (subnetMask[0] & getMyIP[0]) | (~subnetMask[0]);
		  mBCast[1] = (subnetMask[1] & getMyIP[1]) | (~subnetMask[1]);
		  mBCast[2] = (subnetMask[2] & getMyIP[2]) | (~subnetMask[2]);
		  mBCast[3] = (subnetMask[3] & getMyIP[3]) | (~subnetMask[3]);
          mUdp.begin(WiFi.localIP(),mUdpPort);
          mStatus = true;
		  if (mConnAchievedCB != NULL)
			{
				mConnAchievedCB();
			}
          break;
      case SYSTEM_EVENT_STA_DISCONNECTED:
          mStatus = false;
		  if (mConnLostCB != NULL)
			{
				mConnLostCB();
			}
          break;
    }
}
PropWiFiESP32::PropWiFiESP32(char *ssid, char *pwd, byte id, int udpPort)
{
	reference = this;
	mStatus = false;
	mSsidBuf[31] = '\0';
	for (int i = 0; i < 31; i++)
	{
		mSsidBuf[i] = ssid[i];
		if (ssid[i] == '\0')
			break;
	}
	for (int i = 0; i < 31; i++)
	{
		mPwdBuf[i] = pwd[i];
		if (pwd[i] == '\0')
			break;
	}
	mUdpPort = udpPort;
	mId = id;
	mLastConnCheck = 0;
	for(int i = 0; i < 8; i++)
	{
		mMsgQueue[i].idx = 0;
		mHeartbeatTracking[i] = 0;
	}
	mConnAchievedCB = NULL;
	mConnLostCB = NULL;
	mMsgReceivedCB = NULL;
	mMsgAckdCB = NULL;
	mHeartbeatDiscoveredCB = NULL;
	mHeartbeatLostCB = NULL;
	mResendTimer = 0;
	mEnableHeartbeating = false;
	mHeartbeatTimer = 0;
	mHeartbeatListening[0] = 0;
	mHeartbeatListening[1] = 0;
	mHeartbeatListening[2] = 0;
	mHeartbeatListening[3] = 0;
	mHeartbeatListening[4] = 0;
	mHeartbeatListening[5] = 0;
	mHeartbeatListening[6] = 0;
	mHeartbeatListening[7] = 0;
	mMsgIdx = 1;
	mBCast = IPAddress(255,255,255,255);
	WiFi.disconnect(true);
	WiFi.onEvent(WiFiEvent);
	WiFi.begin(mSsidBuf, mPwdBuf);
	mRcvdIdsQueue[0] = 0;
	mRcvdIdsQueue[1] = 0;
	mRcvdIdsQueue[2] = 0;
	mRcvdIdsQueue[3] = 0;
	mRcvdIdsQueue[4] = 0;
	mRcvdIdsQueue[5] = 0;
	mRcvdIdsQueue[6] = 0;
	mRcvdIdsQueue[7] = 0;
	mRcvdIdsQueue[8] = 0;
	mRcvdIdsQueue[9] = 0;
	mRcvdIdsQueue[10] = 0;
	mRcvdIdsQueue[11] = 0;
	mRcvdIdsQueue[12] = 0;
	mRcvdIdsQueue[13] = 0;
	mRcvdIdsQueue[14] = 0;
	mRcvdIdsQueue[15] = 0;
	mRcvdIdsQueue[16] = 0;
	mRcvdIdsQueue[17] = 0;
	mRcvdIdsQueue[18] = 0;
	mRcvdIdsQueue[19] = 0;
	mRcvdIdsIdx = 0;
	//beginConnection();
}
bool PropWiFiESP32::alreadyReceived(byte source, byte msgId)
{
	for (int idx = 0; idx <= 18; idx = idx + 2)
	{
		if ((mRcvdIdsQueue[idx] == source) && (mRcvdIdsQueue[idx + 1] == msgId))
		{
			return true;
		}
	}
	return false;
}
void PropWiFiESP32::markReceived(byte source, byte msgId)
{
	mRcvdIdsQueue[mRcvdIdsIdx] = source;
	mRcvdIdsIdx++;
	mRcvdIdsQueue[mRcvdIdsIdx] = msgId;
	mRcvdIdsIdx++;
	if (mRcvdIdsIdx > 18)
		mRcvdIdsIdx = 0;
}

void PropWiFiESP32::sendHeartbeat()
{
	byte actualMsg[4];
	actualMsg[0] = mId;
	actualMsg[1] = 255;
	actualMsg[2] = 1;
	actualMsg[3] = 0;
	
	mUdp.beginPacket(mBCast, mUdpPort);
	mUdp.write(actualMsg, 4);
	mUdp.endPacket();	
}

void PropWiFiESP32::sendMessageAck(byte recipient, byte msgIdx)
{
	byte actualMsg[5];
	actualMsg[0] = mId;
	actualMsg[1] = recipient;
	actualMsg[2] = 3;
	actualMsg[3] = msgIdx;
	actualMsg[4] = 0;
	
	mUdp.beginPacket(mBCast, mUdpPort);

	mUdp.write(actualMsg, 5);

	mUdp.endPacket();
}

void PropWiFiESP32::sendMessageHelper(byte recipient, byte msgIdx, unsigned long data)
{
	byte actualMsg[9];
	actualMsg[0] = mId;
	actualMsg[1] = recipient;
	actualMsg[2] = 2;
	actualMsg[3] = msgIdx;
	actualMsg[4] = ((byte *)&data)[0];
	actualMsg[5] = ((byte *)&data)[1];
	actualMsg[6] = ((byte *)&data)[2];
	actualMsg[7] = ((byte *)&data)[3];
	actualMsg[8] = 0;
	
	mUdp.beginPacket(mBCast, mUdpPort);
	mUdp.write(actualMsg, 9);
	mUdp.endPacket();
}

void PropWiFiESP32::resendMessages()
{
	unsigned long curMillis = millis();
	for (int i = 0; i < 8; i++)
	{
		if (mMsgQueue[i].idx != 0)
		{
			if (((curMillis < mMsgQueue[i].resent) && (curMillis > propWiFiResendTime)) ||
				(curMillis > mMsgQueue[i].resent + propWiFiResendTime))
			{
				mMsgQueue[i].resent = curMillis;
				sendMessageHelper(mMsgQueue[i].recipient, mMsgQueue[i].idx, mMsgQueue[i].payload);
				if (((curMillis < mMsgQueue[i].sent) && (curMillis > propWiFiMaxResendTime)) ||
					(curMillis > mMsgQueue[i].sent + propWiFiMaxResendTime))
				{
					mMsgQueue[i].idx = 0;
				}
			}
		}
	}
}

void PropWiFiESP32::registerMessageAckdCB(void (*msgAckdCB)(byte msgId))
{
	mMsgAckdCB = msgAckdCB;
}

void PropWiFiESP32::registerMessageReceivedCB(void (*msgReceivedCB)(byte source, unsigned long data))
{
	mMsgReceivedCB = msgReceivedCB;
}

byte PropWiFiESP32::sendMessage(byte recipient, unsigned long data)
{
	int queueIdx = 0;
	for (queueIdx = 0; queueIdx < 8; queueIdx++)
	{
		if (mMsgQueue[queueIdx].idx == 0)
		{
			mMsgQueue[queueIdx].idx = mMsgIdx;
			mMsgQueue[queueIdx].payload = data;
			mMsgQueue[queueIdx].sent = millis();
			mMsgQueue[queueIdx].resent = mMsgQueue[queueIdx].sent;
			mMsgQueue[queueIdx].recipient = recipient;
			break;
		}
	}
	//Actually send message...
	sendMessageHelper(recipient, mMsgIdx, data);
	byte returnVal = mMsgIdx;
	mMsgIdx++;
	if (mMsgIdx == 0)
	{
		mMsgIdx = 1;
	}
	return returnVal;
}




void PropWiFiESP32::registerConnectionAchievedCB(void (*connAchievedCB)())
{
	mConnAchievedCB = connAchievedCB;
}
void PropWiFiESP32::registerConnectionLostCB(void (*connLostCB)())
{
	mConnLostCB = connLostCB;
}



void PropWiFiESP32::processIncoming()
{
	int length = mUdp.parsePacket();
	if ((length > 9) || (length < 4))
	{		
		mUdp.flush();
	}
	else
	{
		byte packetBuff[11];
		packetBuff[7] = 0;
		int actualLength =  mUdp.read(packetBuff, length);
		if ((actualLength == 4) && (packetBuff[2] == 1))
		{
			//Handle heartbeat.
			unsigned long curMillis = 0;
			for (int i = 0; i < 8; i++)
			{
				if (mHeartbeatListening[i] == packetBuff[0])
				{
					if (curMillis == 0)
					{
						curMillis = millis();
						if (curMillis == 0)
							curMillis++;
					}
					if ((mHeartbeatDiscoveredCB != 0) && (mHeartbeatTracking[i] == 0))
					{
						mHeartbeatDiscoveredCB(mHeartbeatListening[i]);
					}
					mHeartbeatTracking[i] = curMillis;
				}
			}
		}
		else if ((actualLength == 5) && (packetBuff[2] == 3))
		{
			//Handle message ack.
			if (packetBuff[1] == mId)
			{
				//This was ack for a message from us...
				for(int i=0; i <  8; i++)
				{
					if (mMsgQueue[i].idx == packetBuff[3])
					{
						mMsgQueue[i].idx = 0;
					}
				}
				if (mMsgAckdCB != NULL)
				{
					mMsgAckdCB(packetBuff[3]);
				}
			}
		}
		else if ((actualLength == 9) && (packetBuff[2] == 2))
		{
			//Handle message.
			//Test if we are destination...
			if (packetBuff[1] == mId)
			{
				//This is a message for us...
				if ((mMsgReceivedCB != NULL) && (alreadyReceived(packetBuff[0], packetBuff[3]) == false))
				{
					unsigned long dataBlock;
					((byte *)&dataBlock)[0] = packetBuff[4];
					((byte *)&dataBlock)[1] = packetBuff[5];
					((byte *)&dataBlock)[2] = packetBuff[6];
					((byte *)&dataBlock)[3] = packetBuff[7];
					mMsgReceivedCB(packetBuff[0], dataBlock);
					markReceived(packetBuff[0], packetBuff[3]);
				}
				sendMessageAck(packetBuff[0], packetBuff[3]);
			}
		}
	}
}
void PropWiFiESP32::registerHeartbeatDiscoveredCB(void (*heartbeatDiscoveredCB)(byte source))
{
	mHeartbeatDiscoveredCB = heartbeatDiscoveredCB;
}

void PropWiFiESP32::registerHeartbeatLostCB(void (*heartbeatLostCB)(byte source))
{
	mHeartbeatLostCB = heartbeatLostCB;
}

void PropWiFiESP32::tick()
{
	unsigned long curMillis = millis();
	if (mStatus)
	{
		processIncoming();
	}
	if (mStatus && ((curMillis > mResendTimer + propWiFiResendTime) ||
		((curMillis < mResendTimer) && (curMillis > propWiFiResendTime))))
	{
		mResendTimer = curMillis;
		resendMessages();
	}
	bool heartbeatTimerUp = false;
	if ((curMillis > mHeartbeatTimer + propWiFiHeartbeatTimer) ||
		((curMillis < mHeartbeatTimer) && (curMillis > propWiFiHeartbeatTimer)))
	{
		mHeartbeatTimer = curMillis;
		if (mStatus && mEnableHeartbeating)
		{
			sendHeartbeat();
		}
		for (int i = 0; i < 8; i++)
		{
			if ((mHeartbeatListening[i] != 0) && (mHeartbeatTracking[i] != 0))
			{
				if (((curMillis < mHeartbeatTracking[i]) && (curMillis > propWiFiHeartbeatTimeout)) ||
					(curMillis > mHeartbeatTracking[i] + propWiFiHeartbeatTimeout))
				{
					mHeartbeatTracking[i] = 0;
					if (mHeartbeatLostCB != NULL)
					{
						mHeartbeatLostCB(mHeartbeatListening[i]);
					}						
				}
			}
		}
	}
}

void PropWiFiESP32::listenHeartbeat(byte source)
{
	for (int i = 0; i < 8; i++)
	{
		if (mHeartbeatListening[i] == 0)
		{
			mHeartbeatListening[i] = source;
			mHeartbeatTracking[i] = 0;
			return;
		}
	}
}
bool PropWiFiESP32::connectionStatus()
{
	return mStatus;
}