/*
  Test.h - Test library for Wiring - description
  Copyright (c) 2006 John Doe.  All right reserved.
*/

// ensure this library description is only included once
#ifndef PropWiFi_h
#define PropWiFi_h

#include <WiFi.h>
#include <WiFiUdp.h>


#define propWiFiResendTime 200
#define propWiFiMaxResendTime 10000
#define propWiFiHeartbeatTimer 600
#define propWiFiHeartbeatTimeout 2500

// library interface description
class PropWiFiESP32
{
  // user-accessible "public" interface
  public:
    PropWiFiESP32(char *ssid, char *pwd, byte id, int udpPort);
    void tick();

	void registerConnectionAchievedCB(void (*connAchievedCB)());
	void registerConnectionLostCB(void (*connLostCB)());
	void registerMessageReceivedCB(void (*msgReceivedCB)(byte source, unsigned long data));
	void registerMessageAckdCB(void (*msgAckdCB)(byte msgId));
	void registerHeartbeatDiscoveredCB(void (*heartbeatDiscoveredCB)(byte source));
	void registerHeartbeatLostCB(void (*heartbeatLostCB)(byte source));
	byte sendMessage(byte recipient, unsigned long data);
	
	void enableHeartbeating(bool state = true);
	
	void listenHeartbeat(byte source);
	
	bool connectionStatus();
	
	void ClassWiFiEvent(WiFiEvent_t event);
  // library-accessible "private" interface
  private:
	
	struct PropWiFiMsgQueueStruct {
		unsigned long sent;
		unsigned long resent;
		byte recipient;
		byte idx;
		unsigned long payload;
	};
  
	PropWiFiMsgQueueStruct mMsgQueue[8];
	byte mRcvdIdsQueue[20];
	byte mRcvdIdsIdx;
	char mSsidBuf[32];
	char mPwdBuf[32];
	byte mId;
	int mUdpPort;
	byte mMsgIdx;
	bool mEnableHeartbeating;
	byte mHeartbeatListening[8];
	WiFiUDP mUdp;
	IPAddress mBCast;
	
	bool mStatus;	
	unsigned long mLastConnCheck;
	unsigned long mResendTimer;
	unsigned long mHeartbeatTimer;
	unsigned long mHeartbeatTracking[8];
	void (*mConnAchievedCB)();
	void (*mConnLostCB)();
	void (*mMsgReceivedCB)(byte source, unsigned long data);
	void (*mMsgAckdCB)(byte msgId);
	void (*mHeartbeatDiscoveredCB)(byte source);
	void (*mHeartbeatLostCB)(byte source);
	bool alreadyReceived(byte source, byte msgId);
	void markReceived(byte source, byte msgId);
	

	void processIncoming();
	void sendMessageHelper(byte recipient, byte msgIdx, unsigned long data);
	void sendMessageAck(byte recipient, byte msgIdx);
	void sendHeartbeat();
	void resendMessages();

};

#endif

